<copyright> MU tom doc lexer.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright (C) 1996 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: LexDoc.t,v 1.6 1998/04/26 22:47:05 tiggr Exp $</id>
    </copyright>

implementation class
LexDoc: SkeletonLexer, ParseDocDefines, mu

end;

implementation instance
LexDoc
{
  <doc> The string we are lexing.  </doc>
  String string;

  <doc> The next character to lex.  </doc>
  int next;

  <doc> The cached length of the {string}.  </doc>
  int length;
}

<doc> This is the designated initializer.  It may be called repeatedly to
    re-initialize this lexer.  </doc>
id
  initWithFilename String name
	      line int line
	       doc String doc
{
  /* Freeze the string so we get fast substrings of the constant string,
     instead of slow substrings of a (possibly) non-constant string. */
  (string, next) = ([doc frozen], 0);
  length = [string length];

  = [super initWithFilename name line: line];
}

<doc> Return the next token.  </doc>
int
  nextToken
{
  if (next == length)
    {
      semantics = nil;
      return token = 0;
    }

  int start = next;
  char c = string[next++];

  if (c == '{')
    {
      int nesting = 1;

      start = next;
      token = IDENTIFIER;
      while (next < length && nesting != 0)
	{
	  c = string[next++];
	  if (c == '{')
	    {
	      token = CODE;
	      nesting++;
	    }
	  else if (c == '}')
	    nesting--;
	  else if (c == '\n')
	    {
	      token = CODE;
	      current_line++;
	    }
	  else if (token == IDENTIFIER
		   && !((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
			|| (c >= '0' && c <= '9') || c == '-' || c == '_'))
	    token = CODE;
	}

      if (c != '}')
	{
	  [[[Parser error] print "unterminated `{' in documentation"] nl];
	  token = TEXT;
	}

      semantics = [string substring (start, next - start - 1)];
    }
  else
    {
      /* Search for the end of the doc, a `{', or a double new-line.  */
      while (next < length && c != '{')
	{
	  if (c == '\n')
	    if (next == length)
	      current_line++;
	    else
	      {
		char d = string[next];

		if (d == '\n')
		  {
		    if (next - start > 1)
		      {
			next--;
			break;
		      }

		    /* Eat as many new-lines as possible.  */
		    do
		      {
			current_line++;
			c = string[next++];
		      } while (c == '\n');

		    next--;
		    semantics = nil;
		    return token = PAR;
		  }
	      }
	  c = string[next++];
	}

      if (c == '{')
	next--;

      semantics = [string substring (start, next - start)];
      token = TEXT;
    }

  = token;
}

end;
