<copyright> MU meta.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: Meta.t,v 1.21 1998/07/27 15:20:57 tiggr Exp $</id>
    </copyright>

/******************** Meta ********************/

implementation class
Meta: TypeSingle
{
  <doc> The search mark for searching through the hierarchy.  </doc>
  static int search_mark;

  <doc> All class metas.  </doc>
  public static MutableArray all_class_metas;

  <doc> Similarly, all instance metas.  </doc>
  public static MutableArray all_instance_metas;
}

<doc> Designated allocator.  Create new {Class} and {Instance} meta
    objects for the class with the {name} contained in the {file}.  </doc>
instance (id)
  withName String name
      file TomFile f
{
  if (!all_class_metas)
    {
      all_class_metas = [MutableObjectArray withCapacity 128];
      all_instance_metas = [MutableObjectArray withCapacity 128];
    }

  Instance i = [[Instance alloc] initWithName name];
  [Extension ofMeta i inFile f];
  [all_instance_metas add i];

  Class c = [[Class alloc] initWithName name instance i];
  [Extension ofMeta c inFile f];
  [all_class_metas add i];

  [i set_meta c];
  [c set_meta meta_meta];

  = i;
}

<doc> Initialize the {Meta} class.  The unit tom has already been loaded.
    Does nothing if {bare_environment}.  Primary action is to set the meta
    classes.  </doc>
void
  initialize boolean bare_environment
{
  if (!bare_environment)
    {
      Unit u = [Unit named UNIT_NAME_TOM];
      Mapped metas = [u metas];
      Enumerator e = [metas enumerator];
      Instance i;
      boolean b;

      meta_meta = metas[META_NAME_STATE];
      while ({(b, i) = [e next]; b;})
	{
	  // Why is this cast necessary?
	  // Tue Jul 15 15:52:11 1997, tiggr@natlab.research.philips.com
	  [Class ([i meta]) set_meta meta_meta];
	}
    }
}

<doc> Return the {Meta} referred to by the name {unit}.{meta} (where
    {unit} is {nil} to indicate absence of the unit indication), as seen
    from the {file}.  This takes into account the units visible from that
    file.  </doc>
instance (id)
   named (String, String) (unit, meta)
  inFile TomFile f
{
  if (!unit)
    = [[f unit] metaNamed meta recurse: YES];
  else
    {
      Unit u = [Unit named unit];
  
      if (!u)
	= nil;
      else
  	= [u metaNamed meta];
    }
}

end;

implementation instance
Meta
{
  <doc> The class of this meta.  </doc>
  public mutable Class meta;

  <doc> The main extension.  </doc>
  mu.Extension main;

  <doc> Other extensions defined for this meta, keyed on their name.
      </doc>
  MutableDictionary extensions;

  <doc> All extensions, starting with the main extension, in an easy
      traversable indexed collection.  </doc>
  MutableArray all_extensions;

  <doc> The variables and constants in any of the extensions of this meta,
      keyed on their name.  </doc>
  MutableDictionary attributes;

  <doc> Behaviour supers.  </doc>
  public MutableIndexed behaviour_supers;

  <doc> State supers.  </doc>
  public MutableIndexed state_supers;

  <doc> Behaviour subs.  </doc>
  public MutableIndexed behaviour_subs;

  <doc> State subs.  </doc>
  public MutableIndexed state_subs;

  <doc> The meta as which we pose.  </doc>
  public Meta posed;

  <doc> The meta by which we are posed.  </doc>
  public Meta poser;

  <doc> The value of {search_mark} the last time this meta was visited.
      </doc>
  int mark;
}

<doc> Designated initializer.  </doc>
id
  initWithName String name
{
  all_extensions = [MutableObjectArray new];
  = [super initWithName name];
}

<doc> Return the {class} of the receiving instance.  The default
    implementation refuses to do anything.  It should be usefully
    implemented by the {Instance} class.  </doc>
Class
  itsClass
{
  [self shouldNotImplement cmd];
}

<doc> Return the unit containing (the main extension of) this class.  </doc>
Unit
  unit
{
  = [[self extensionNamed nil] unit];
}

/********** inheritance **********/

<doc><h4>Inheritance</h4></doc>

<doc> Add the meta {m} to the behaviour supers.  This tells {m} to add the
    receiving meta to its behaviours subs.  </doc>
void
  addBehaviourSuper Meta m
{
  if (!behaviour_supers)
    behaviour_supers = [MutableObjectArray new];
  [behaviour_supers add m];
  [m addBehaviourSub self];
}

<doc> Note the meta {m} to be a behavioural subclass of the receiving
    meta.  </doc>
protected void
  addBehaviourSub Meta m
{
  if (!behaviour_subs)
    behaviour_subs = [MutableObjectArray new];
  [behaviour_subs add m];
}

<doc> Add the {meta} to the state supers.  This tells {m} to add the
    receiving meta to its state subs.  </doc>
void
  addStateSuper Meta m
{
  if (!state_supers)
    state_supers = [MutableObjectArray new];
  [state_supers add m];
  [m addStateSub self];
}

<doc> Note the meta {m} to be a subclass of the receiving meta.  </doc>
protected void
  addStateSub Meta m
{
  if (!state_subs)
    state_subs = [MutableObjectArray new];
  [state_subs add m];
}

<doc> Return the posing meta ultimately referenced when referencing
    the receiving meta.  </doc>
id
  posedSelf
{
  = !poser ? self : [poser posedSelf];
}

<doc> Note that the receiving meta is posing as {p}.  </doc>
protected void
  setPosed Meta p
pre
  !posed
{
  posed = p;
}

<doc> Note the meta {p} to pose the receiving meta.  </doc>
void
  setPoser Meta p
pre
  [self isInstance] == [p isInstance]
{
  if (poser != nil)
    [poser setPoser p];
  else
    {
      poser = p;
      [p setPosed self];
    }
}

/********** extensions **********/

<doc><h4>Extensions</h4></doc>

<doc> Add the {Extension} {x} to the {extensions} of this meta.  An error
    is reported (and the extension is not added) if an extension with the
    same name already exists.  </doc>
void
  addExtension Extension x
{
  String nx = [x name];

  if (!nx)
    if (!main)
      all_extensions[0] = main = x;
    else
      [[[Parser error] print "attempt to redefine main extension"] nl];
  else if (extensions != nil && extensions[nx] != nil)
    [[[Parser error] print ("attempt to redefine extension ",
			    [x descriptive_name])] nl];
  else
    {
      if (!extensions)
	{
	  extensions = [MutableDictionary new];
	  if (![all_extensions length])
	    [all_extensions add nil];
	}
      extensions[nx] = x;
      [all_extensions add x];
    }
}

<doc> Return the {Extension} with the {name}.  Return the {main} extension
    if {name} is {nil}.  </doc>
Extension
  extensionNamed String name
{
  = !name ? main : !extensions ? nil : extensions[name];
}

/********** attributes **********/

<doc><h4>Attributes</h4></doc>

<doc> Note the addition of the alias {a} to the extension {e}.  </doc>
void
  alias Alias a
     in Extension e
pre
  self == [e meta] && [a isAlias]
{
  String aname = [a name];

  // Check this, check that...
  // Sun Apr 26 18:29:26 1998, tiggr@gerbil.org

  if (!attributes)
    attributes = [MutableDictionary new];
  attributes[aname] = a;
}

<doc> Note the addition of the constant {c} to the extension {e}.  </doc>
void
  constant Constant c
	in Extension e
pre
  self == [e meta]
{
  String cname = [c name];

  // Check this, check that...
  // Sun Apr 26 18:29:26 1998, tiggr@gerbil.org

  if (!attributes)
    attributes = [MutableDictionary new];
  attributes[cname] = c;
}

<doc> Note the addition of the variable {v} to the extension {e}.  </doc>
void
  variable MetaVariable v
	in Extension e
pre
  self == [e meta] && ![v isAlias]
{
  String vname = [v name];

  if ([v isStatic] && ![self isClass])
    {
      // Clear staticness...
      // Sat Apr 25 15:12:52 1998, tiggr@gerbil.org
      [[[Parser error v]
	 print ("can't add static `", vname, "' to ", self)] nl];
    }

  // This is not fully correct.
  // Sun Apr 26 17:53:37 1998, tiggr@gerbil.org
  Attribute o = [self attributeNamed vname
		      including (TRUE, TRUE, FALSE, TRUE, TRUE, FALSE)];
  if (o != nil)
    {
      /* An attribute declaration with the same name is not allowed within
         the same unit (and thus also not within the same extension).  */
      if ([[e file] unit] == [[[o extension] file] unit])
	[[[Parser error v] print ("declaration of `", vname,
				  "' conflicts with previous declaration")] nl];
      else
	[[[Parser warning v] print ("declaration of `", vname,
				    "' hides previous declaration")] nl];
      [[[Parser report o] print ("`", vname, "' previously declared here")] nl];
    }

  /* A non-main extension overrides a decl from the main extension.  */
  if (!o || [[o extension] meta] != self || ![e isMainExtension])
    {
      if (!attributes)
	attributes = [MutableDictionary new];
      attributes[vname] = v;
    }
}

<doc> Look for a meta attribute with the {name}.  Iff {constants}, const
    thingies are acceptable.  Iff {originals}, true meta variable
    declarations are accepted.  Iff {aliases}, redeclarations of meta
    variables are sought.  Iff {supers}, super classes may aid to their
    best.  Iff {do_self}, also the receiving meta may do its best.  Iff
    {do_class}, searching continues with the class of the receiving meta
    to satisfy the query, if needed.  </doc>
Attribute
  attributeNamed String name
       including (boolean, boolean, boolean, boolean, boolean, boolean)
		 (constants, originals, aliases, supers, do_self, do_class)
      excluding: Attribute to_skip = nil
{
  boolean multiple;
  Attribute v;

  (multiple, v) = [self attributeNamed name
			including (constants, originals, aliases,
				   do_self, supers)
			excluding to_skip mark ++search_mark];
  if (!multiple && !v && do_class && meta != nil)
    (multiple, v) = [meta attributeNamed name
			  including (constants, originals, aliases,
				     do_self, supers)
			  excluding to_skip mark search_mark];

  = multiple ? nil : v;
}

// This searching should be eliminated by maintaining information on the
// flattened instance variables in each meta.
// Wed Jan 22 14:59:46 1997, tiggr@akebono.ics.ele.tue.nl
protected (boolean, Attribute)
  attributeNamed String name
       including (boolean, boolean, boolean, boolean, boolean)
		 (constants, originals, aliases, do_self, do_supers)
       excluding Attribute to_skip
	    mark int k
pre
  /* Ensure we're never visited while already being traversed.  */
  mark <= k
{
  if (mark == k)
    return (FALSE, nil);
  mark = k;

  Attribute v;

  if (do_self && attributes != nil)
    {
      v = attributes[name];
      if (v != nil && v != to_skip)
	return (FALSE, v);

      /* Make sure V is nil; it could be TO_SKIP.  */
      v = nil;
    }

  if (do_supers && state_supers != nil)
    {
      int i, n = [state_supers length];

      for (i = 0; i < n; i++)
	{
	  Meta m = state_supers[i];
	  boolean multiple;
	  Attribute w;

	  (multiple, w) = [m attributeNamed name
			     including (constants, originals, aliases,
					TRUE, do_supers)
			     excluding to_skip mark k];
	  if (multiple || (w != nil && v != nil))
	    return (TRUE, nil);

	  if (w != nil)
	    v = w;
	}
    }

  // Again!?
  if (do_supers && behaviour_supers != nil)
    {
      int i, n = [behaviour_supers length];

      for (i = 0; i < n; i++)
	{
	  Meta m = behaviour_supers[i];
	  boolean multiple;
	  Attribute w;

	  (multiple, w) = [m attributeNamed name
			     including (constants, originals, aliases,
					TRUE, do_supers)
			     excluding to_skip mark k];
	  if (multiple || (w != nil && v != nil))
	    return (TRUE, nil);

	  if (w != nil)
	    v = w;
	}
    }

  return (FALSE, v);
}

/********** miscellaneous **********/

<doc><h4>Miscellaneous</h4></doc>

boolean
  isClass
{
  = NO;
}

boolean
  isInstance
{
  = NO;
}

redefine boolean
  isMeta
{
  = YES;
}

end;

/******************** Class ********************/

implementation class
Class: Meta

end;

implementation instance
Class
{
  <doc> The instance of this class, i.e. the meta in the other direction.
      </doc>
  Instance etam;
}

<doc> Add the {meta} to the state supers, and also with our instance.  </doc>
void
  addStateSuper Meta m
{
  [super addStateSuper [m meta]];
  [etam addSingleStateSuper m];
}

<doc> Add the {meta} only to our state supers, not with our instance.  </doc>
void
  addSingleStateSuper Meta m
{
  [super addStateSuper m];
}

<doc> Designated initializer.  </doc>
id
  initWithName String name
      instance Instance i
{
  etam = i;
  = [super initWithName name];
}

<doc> Return {TRUE}.  </doc>
redefine boolean
  isClass
{
  = YES;
}

<doc> Return the instance of this class.  </doc>
redefine Instance
  instance
{
  = etam;
}

end;

/******************** Instance ********************/

implementation class
Instance: Meta
end;

implementation instance
Instance

<doc> Add the {meta} to the state supers, and also with our class.  </doc>
redefine void
  addStateSuper Meta m
{
  [super addStateSuper m];
  [Class (meta) addSingleStateSuper [m meta]];
}

<doc> Add the {meta} only to our state supers, not with our class.  </doc>
void
  addSingleStateSuper Meta m
{
  [super addStateSuper m];
}

<doc> Issue an error, since instances do not have instances (only classes
    do).  </doc>
redefine Instance (self)
  instance
{
  [[[Parser error] print "instances do not have instances"] nl];
}

<doc> Return {TRUE}.  </doc>
redefine boolean
  isInstance
{
  = YES;
}

<doc> Return the {meta}.  </doc>
Class
  itsClass
{
  = meta;
}

<doc> Pass to {super}, followed by our {meta}.  </doc>
void
  setPoser Instance p
{
  [super setPoser p];
  [meta setPoser [p itsClass]];
}

end;
