<copyright> MU global defines.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright (C) 1996 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: mu.t,v 1.28 1999/08/13 21:19:31 tiggr Exp $</id>
    </copyright>

implementation class
mu: instance (All)
{
  <doc> Important names.  </doc>
  const UNIT_NAME_TOM = "tom";
  const META_NAME_ANY = "Any";
  const META_NAME_TOP = "Top";
  const META_NAME_STATE = "State";
  const STRING_CST_CLASS = "ByteString";
  const CONDITION_CLASS = "Condition";
  const BLOCK_CLASS = "Block";

  <doc> The basic types.  </doc>
  const BT_FIRST = 0;
  const BT_VOID = 0;
  const BT_BOOLEAN = 1;
  const BT_BYTE = 2;
  const BT_CHAR = 3;
  const BT_INT = 4;
  const BT_LONG = 5;
  const BT_FLOAT = 6;
  const BT_DOUBLE = 7;
  const BT_POINTER = 8;
  const BT_SELECTOR = 9;
  const BT_DYNAMIC = 10;
  const BT_ID = 11;
  const BT_META_ID = 12;
  const BT_ETAM_ID = 13;
  const BT_BEYOND = 14;

  <doc> Mapping from the basic kind (one of {BT_*}) to the {BasicType}
      object.  </doc>
  static Indexed basic_types_from_kind;

  <doc> Closed naming of the operators, in non-decreasing priority.  </doc>
  const OP_ASSIGN = 0;
  const OP_ITE = 1;
  const OP_IMPLIES = 2;
  const OP_SC_OR = 3;
  const OP_SC_AND = 4;
  const OP_GT = 5;
  const OP_GE = 6;
  const OP_EQ = 7;
  const OP_NE = 8;
  const OP_LE = 9;
  const OP_LT = 10;
  const OP_EOR = 11;
  const OP_OR = 12;
  const OP_AND = 13;
  const OP_SHR = 14;
  const OP_SHL = 15;
  const OP_LSR = 16;
  const OP_SUB = 17;
  const OP_ADD = 18;
  const OP_MOD = 19;
  const OP_DIV = 20;
  const OP_MUL = 21;
  const OP_NOT = 22;
  const OP_INV = 23;
  const OP_MIN = 24;
  const OP_PP = 25;
  const OP_MM = 26;
  const OP_OLD = 27;
  const OP_SET = 28;
  const OP_RET = 29;
  const OP_AT = 30;
  
  <doc> Indexed on the above, the operator name strings.  </doc>
  static Indexed operator_names;

  <doc> The error type.  </doc>
  static ErrorType error_type;

  <doc> The set of type names.  </doc>
  static MutableSet types;

  <doc> The builtin types.  </doc>
  static MutableDictionary basic_types;

  <doc> The dynamic type (one of the {basic_types}).  </doc>
  static BasicType dynamic_type;

  <doc> The any type.  </doc>
  static AnyType any_type;

  <doc> The type of all objects (both classes and instances).  </doc>
  static Meta top_type;

  <doc> The type that is all objects (both classes and instances).  </doc>
  static Meta bottom_type;

  <doc> The void basic type.  </doc>
  static BasicType void_type;

  <doc> The selector type.  </doc>
  static BasicType selector_type;

  <doc> The type of string constants.  </doc>
  static Meta string_cst_type;

  <doc> The type of conditions.  </doc>
  static Meta condition_type;

  <doc> The state class, which is the class object of every class.  </doc>
  static Class meta_meta;

  <doc> The expression with the {top_type}, used as the receiver of all
      ordinary operator invocations.  </doc>
  static Expression top_expression;

  // This information is not used yet, since traversing the unit tree to
  // find a meta is about as costly as traversing it to check if the unit
  // containing a meta is within the scope of the referring unit.
  // Mon Sep  2 12:56:43 1996, tiggr@cobra.es.ele.tue.nl
  <doc> The metas which only exist in a single unit.  </doc>
  static MutableDictionary unique_metas;

  <doc> The error expression.  </doc>
  static ErrorExpression error_expr;

  <doc> The expression with the void type.  </doc>
  static Expression void_expr;

  <doc> All units, in a topological sort.  </doc>
  static MutableArray all_user_units;

  <doc> All units, including the builtin one.  </doc>
  static MutableArray all_units;

  <doc> The TOM unit.  </doc>
  static Unit unit_tom;
}

<doc> Initialize everything in the mu unit.  If {bare_environment} is
    {TRUE}, no preloading of the TOM unit will be performed.  </doc>
void
  init boolean bare_environment
{
  operator_names = [ObjectArray with ("=", "? :", "->", "||", "&&",
				      ">", ">=", "==", "!=", "<=",
				      "<", "^", "|", "&", ">>",
				      "<<", ">>>", "-", "+", "%", "/",
				      "*", "!", "~", "-", "++",
				      "--", "old", "=", "return", "[]")];
  error_type = [ErrorType new];
  error_expr = [ErrorExpression new];

  unique_metas = [MutableDictionary new];
  basic_types = [MutableDictionary new];

  Unit u = [Unit builtin_unit];
  all_units = [MutableObjectArray with u];
  all_user_units = [MutableObjectArray new];

  TomFile f = [TomFile builtin_file];
  bottom_type = [Meta withName META_NAME_ANY file f];
  [f addElement [bottom_type extensionNamed nil]];
  [u addMeta bottom_type];
  top_type = [Meta withName META_NAME_TOP file f];
  [f addElement [top_type extensionNamed nil]];
  [u addMeta top_type];
  [bottom_type addStateSuper top_type];
  any_type = [AnyType shared];
  top_expression = [[Expression alloc] initWithType top_type];

  if (!bare_environment)
    {
      unit_tom = [Unit named UNIT_NAME_TOM load: YES];
      if (!unit_tom)
	{
	  [[[Parser error] print ("Unable to load unit `", UNIT_NAME_TOM, '\'')]
	    nl];
	  return;
	}
    }

  [Meta initialize bare_environment];
  [BasicType initialize bare_environment];

  selector_type = basic_types["selector"];
  void_type = basic_types["void"];
  void_expr = [[Expression alloc] initWithType void_type];
  if (bare_environment)
    {
      /* Must have *some* type to fill in for a condition.  */
      condition_type = top_type;
    }
  else
    {
      string_cst_type = [unit_tom metaNamed STRING_CST_CLASS];
      if (!string_cst_type)
	[[[Parser error] print ("unknown class ", UNIT_NAME_TOM,
				".", STRING_CST_CLASS)] nl];
      condition_type = [unit_tom metaNamed CONDITION_CLASS];
      if (!condition_type)
	[[[Parser error] print ("unknown class ", UNIT_NAME_TOM,
				".", CONDITION_CLASS)] nl];
    }
}

end;

implementation instance
mu: instance (All)
end;
