<copyright> MU tom semantics.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1996, 1997 Pieter J. Schoenmakers.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: semantics.t,v 1.48 1999/09/25 21:29:29 tiggr Exp $</id>
    </copyright>

<doc> This file contains all kinds of classes providing a skeleton for
    behaviour needed by the {ParseTom} tom parser.  </doc>

/******************** Alias ********************/

<doc> An alias is a redeclaration of a {MetaVariable}.  </doc>
implementation class
Alias: StateAttribute

end;

implementation instance
Alias
{
  <doc> The original.  </doc>
  public MetaVariable original;
}

<doc> Return {TRUE}.  </doc>
boolean
  isAlias
{
  = TRUE;
}

<doc> Be informed of the original of this alias.  </doc>
void
  set_original Attribute variable
pre
  !original && [variable isVariable]
{
  original = MetaVariable (variable);
}

<doc> Tell the extension about the receiving alias.  </doc>
void
  register
{
  [ext addAlias self];
}

end;

/******************** AnyExpression ********************/

implementation class
AnyExpression: Expression, AnyThing

end;

implementation instance
AnyExpression

<doc> Designated allocator.  </doc>
id
  init
{
  = [super initWithType [AnyType shared]];
}

end;

/******************** AnyThing ********************/

implementation class
AnyThing: mu

end;

implementation instance
AnyThing

end;

/******************** AnyType ********************/

implementation class
AnyType: Type, AnyThing
{
  <doc> The shared any type.  </doc>
  static AnyType shared;
}

<doc> Designated allocator.  </doc>
instance (id)
  shared
{
  if (!shared)
    shared = [self new];
  = shared;
}

end;

implementation instance
AnyType

end;

/******************** Argument ********************/

<doc> An argument is a parameter or tuple of parameters to a method.  </doc>
implementation class
Argument: Variable
end;

implementation instance
Argument
{
  <doc> Default argument value.  </doc>
  public mutable Expression default_value;
}

deferred void
  feedArgumentSinglesTo Callable m;

end;

/******************** ArgumentSingle ********************/

implementation class
ArgumentSingle: VariableSingle, Argument
end;

implementation instance
ArgumentSingle
{
  <doc> The callable to which this argument... is an argument.  </doc>
  public Callable callable;
}

id
  initWithName String name
	  type Type type
      callable Callable c
{
  callable = c;
  = [super initWithName name type type];
}

void
  feedArgumentSinglesTo Callable m
pre
  !callable
{
  callable = m;
  [m addArgumentSingle self];
}

end;

/******************** ArgumentTuple ********************/

implementation class
ArgumentTuple: VariableTuple, Argument
end;

implementation instance
ArgumentTuple

void
  feedArgumentSinglesTo Callable m
{
  int i;

  while (i < length)
    [self[i++] feedArgumentSinglesTo m];
}

void
  resolveReturnNamesInMethod CustomMethod m
{
  int i;

  for (i = 0; i < length; i++)
    self[i] = [self[i] resolveReturnNamesInMethod m];
}

ArgumentSingle
  returnValueNamed String name
{
  int i;

  for (i = 0; i < length; i++)
    {
      ArgumentSingle rv = [self[i] returnValueNamed name];
      if (rv != nil)
	return rv;
    }
}

end;

/******************** ArrayReference ********************/

implementation class
ArrayReference: Expression

end;

implementation instance
ArrayReference
{
  <doc> The array being indexed.  </doc>
  Expression array;

  <doc> The index.  </doc>
  Expression index;
}

<doc> Designated initializer.  </doc>
id
  initWithArray Expression a
	  index Expression i
{
  (array, index) = (a, i);
  = [super initWithType nil];
}

<doc> Return {TRUE}.  </doc>
boolean
  isArrayReference
{
  = TRUE;
}

<doc> Emit the {referenced}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" array=", array, " index=", index)];
}

end;

/******************** Attribute ********************/

implementation class
Attribute: Common

end;

implementation instance
Attribute
{
  <doc> The extension containing the declaration of this attribute.
      </doc>
  Extension ext;
}

<doc> Designated initializer.  Does not invoke the initializer of {super}.
    </doc>
id
  initWithExtension Extension e
{
  ext = e;
  = self;
}

<doc> Return the extension {ext}.  </doc>
Extension
  extension
{
  = ext;
}

<doc> Return {TRUE} iff the receiving attribute is an alias.  </doc>
boolean
  isAlias
{
}

<doc> Return {TRUE} iff the receiving attribute is a variable.  </doc>
boolean
  isVariable
{
}

<doc> Return the {meta} of the extension {ext}.  </doc>
Meta
  meta
{
  = [ext meta];
}

<doc> Return the name of the receiving attribute.  </doc>
deferred String
  name;

<doc> Register the receiving attribute with its extension.  </doc>
deferred void
  register;

end;

/******************** BasicType ********************/

implementation class
BasicType: TypeSingle

<doc> Create a new basic type and add it to the {basic_types}.  </doc>
protected instance (id) (t)
  createBasicType (String, int) (name, kind)
{
  name = [UniqueString with name];
  t = [[self alloc] initWithName name kind kind];
  basic_types[name] = t;
}

<doc> Initialize the {BasicType} class.  Note that this is _not_ invoked by
    the runtime, but by {mu}.  </doc>
void
  initialize boolean bare_environment
{
  dynamic_type = [self createBasicType ("dynamic", BT_DYNAMIC)];
  basic_types_from_kind = [ObjectArray with
			    ([self createBasicType ("void", BT_VOID)],
			     [self createBasicType ("boolean", BT_BOOLEAN)],
			     [self createBasicType ("byte", BT_BYTE)],
			     [self createBasicType ("char", BT_CHAR)],
			     [self createBasicType ("int", BT_INT)],
			     [self createBasicType ("long", BT_LONG)],
			     [self createBasicType ("float", BT_FLOAT)],
			     [self createBasicType ("double", BT_DOUBLE)],
			     [self createBasicType ("pointer", BT_POINTER)],
			     [self createBasicType ("selector", BT_SELECTOR)],
			     dynamic_type,
			     [self createBasicType ("id", BT_ID)],
			     [self createBasicType ("class (id)", BT_META_ID)],
			     [self createBasicType ("instance (id)",
						    BT_ETAM_ID)])];
}

<doc> Add the basic types to the {set}.  </doc>
void
  addTypes MutableSet set
{
  int i;
  for (i = BT_FIRST; i < BT_BEYOND; i++)
    [set add [basic_types_from_kind[i] name]];
}

end;

implementation instance
BasicType
{
  <doc> This basic type's kind.  </doc>
  public int kind;
}

<doc> Designated initializer.  </doc>
protected id
  initWithName String n
	  kind int k
{
  kind = k;

  = [super initWithName n];
}

<doc> Return {BT_ID} in case we are {BT_ETAM_ID}.  </doc>
id
  asSeenFromInstance Instance i
{
  = kind == BT_ETAM_ID ? basic_types_from_kind[BT_ID] : self;
}

Type
  instance
{
  = self;
  if (kind == BT_META_ID)
    = basic_types_from_kind[BT_ID];
  else if (kind == BT_ID)
    = basic_types_from_kind[BT_ETAM_ID];
  else if (kind == BT_ETAM_ID)
    [[[Parser error] print "instances do not have instances"] nl];
  else 
    [[[Parser error] print "basic types do not have instances"] nl];
}

redefine boolean
  isBasicType
{
  = YES;
}

Type
  meta
{
  if (kind == BT_ID)
    return basic_types_from_kind[BT_META_ID];
  if (kind == BT_ETAM_ID)
    return basic_types_from_kind[BT_ID];
  if (kind == BT_META_ID)
    [[[Parser error] print "nonsense to ask a class for its class"] nl];
  else 
    [[[Parser error] print "basic types do not have a class"] nl];
  = self;
}

end;

/******************** BlockVariable ********************/

implementation class
BlockVariable: VariableSingle

end;

implementation instance
BlockVariable
{
  <doc> The block to which this variable belongs.  </doc>
  BlockDefinition block;

  <doc> Initial value of this variable.  </doc>
  public Expression initial_value;
}

<doc> Designated initializer.  </doc>
id
  initWithName String n
	  type Type t
	 value Expression e
	 block BlockDefinition b
{
  (block, initial_value) = (b, e);
  = [super initWithName n type t];
}

end;

/******************** ClassReference ********************/

implementation class
ClassReference: Expression

end;

implementation instance
ClassReference

<doc> Designated initializer.  </doc>
id
  initWithClass Class c
{
  = [super initWithType c];
}

end;

/******************** Cast ********************/

implementation class
Cast: Expression

<doc> Designated allocator.  </doc>
instance (id)
  cast Expression e
    to Type t
{
  = [[self alloc] initWithDesiredType t expression e];
}

end;

implementation instance
Cast
{
  <doc> The type of the result.  </doc>
  Type desired_type;

  <doc> The expression to fit.  </doc>
  Expression expr;
}

<doc> Designated initializer.  </doc>
protected id
  initWithDesiredType Type t
	   expression Expression e
{
  desired_type = t;
  expr = e;
  = [super initWithType nil];
}

<doc> Emit the {expr}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" expr=", expr)];
}

end;

/******************** Common ********************/

implementation class
Common: State, FileElement, Conditions, mu

<doc> Designated allocator, overridden to invoke {updateLocation} on the
    returned object.  Direct subclasses of {Common} do not need to do a
    {[super init]}, which presents significant speedup.  </doc>
redefine instance (id) (new_object)
  alloc
{
  new_object = [super (class (State)) alloc];
  [new_object updateLocation];
}

<doc> Similar to {alloc}, but set the location information of the
    {new_object} from the argument {object}.  If {object} is {nil}, this
    performs the same function as (argumentless) {alloc}.  </doc>
instance (id) (new_object)
  alloc Common object
{
  new_object = [super (class (State)) alloc];
  if (!object)
    [new_object updateLocation];
  else
    [new_object updateLocation object];
}

end;

implementation instance
Common
{
  <doc> The file containing this thing.  </doc>
  public mutable String containing_file;

  <doc> The associated line number.  </doc>
  public mutable int line;
}

<doc> Update the location information from (the lexer of) the current
    parser.  </doc>
void
  updateLocation
{
  Parser p = [Parser current_parser];
  if (p != nil)
    {
      Lexer l = Lexer ([p lex]);
      containing_file = [l filename];
      line = [l current_line];
    }
}

<doc> Update the location from the {other}.  </doc>
void
  updateLocation Common other
{
  (containing_file, line) = ([other containing_file], [other line]);
}

<doc> Emit the {containing_file} and {line}.    </doc>
OutputStream
  describeLocation OutputStream s
	       for Parser parser
{
  = [s print (containing_file, ':', line)];
}

<doc> Only errors are errors.  </doc>
boolean
  isError
{
  = NO;
}

<doc> Only tuples are tuples.  </doc>
boolean
  isTuple
{
  = NO;
}

<doc> Return the number of non-tuple elements in this object.  Default
    implementation returns 1.  </doc>
int
  numFlatElements
{
  = 1;
}

<doc> Describe the location, if it is not nil.  </doc>
OutputStream
  writeFields OutputStream s
{
  s = [super writeFields s];
  = (!containing_file ? s
     : [s print (" ", [File basename containing_file], ":", line)]);
}

end;

/******************** Expression ********************/

<doc> An {Expression} is something with a (deducable) type.  </doc>
implementation class
Expression: Typed
end;

implementation instance
Expression

<doc> Return this expression again.  </doc>
id
  again
{
  = self;
}

<doc> Return {TRUE} if the receiving expression is an array reference.
    The default implementation returns {FALSE}.  </doc>
boolean
  isArrayReference
{
}

<doc> Return {TRUE} if the receiving expression is a constant string.
    The default implementation returns {FALSE}.  </doc>
boolean
  isStringConstant
{
}

end;

/******************** ExpressionTuple ********************/

implementation class
ExpressionTuple: Expression, Tuple
end;

implementation instance
ExpressionTuple

// Pity we can't direct this straight to the Tuple...
// Fri Apr 24 14:43:42 1998, tiggr@gerbil.org
<doc> Return {YES}.  </doc>
boolean
  isTuple
{
  = YES;
}

<doc> Direct at the {Tuple}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [super (Tuple) writeFields s];
}

end;

/******************** LocalVariable ********************/

implementation class
LocalVariable: VariableSingle
end;

implementation instance
LocalVariable
{
  <doc> The compound containing this declaration.  </doc>
  public mutable Compound container;
}

<doc> Designated initializer.  </doc>
id
  initWithName String n
	  type Type t
     container Compound c
{
  container = c;
  = [super initWithName n type t];
}

end;

/******************** MetaVariable ********************/

implementation class
MetaVariable: StateAttribute, Qualified
end;

implementation instance
MetaVariable

<doc> Designated initializer.  </doc>
id
  initWithName String n
    qualifiers int q
	  type Type t
     extension Extension e
{
  [super initWithExtension e];
  [super initWithQualifiers q];
  = [super initWithName n type t];
}

<doc> Return {TRUE}.  </doc>
boolean
  isVariable
{
  = TRUE;
}

<doc> Tell the extension about the receiving meta variable.  </doc>
void
  register
{
  [ext addVariable self];
}

end;

/******************** Named ********************/

<doc> A {Named} entity is something with a name.  </doc>
implementation class
Named: Common

end;

implementation instance
Named
{
  <doc> Our name.  </doc>
  public String name;
}

<doc> Designated initializer.  </doc>
id (self)
  initWithName String n
{
  name = n;
}

<doc> Return {self} if the {name} matches {n}.  Used by {LocalVariable}
    and {ArgumentSingle} when naming (the elements of) the return value.
    </doc>
id
  returnValueNamed String n
{
  if ([n equal name])
    = self;
}

<doc> Emit the {name}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" name=", name)];
}

end;

/******************** Qualified ********************/

<doc> A {Qualified} is a {Common} with qualifiers.  </doc>
implementation class
Qualified: Common, ParseTomDefines

<doc> Return the value of the extern bit in the {qualifiers}.  </doc>
boolean
  isExtern int qualifiers
{
  = qualifiers & 1 << (EXTERN - EXTERN) != 0;
}

boolean
  isLocal int qualifiers
{
  = qualifiers & 1 << (LOCAL - EXTERN) != 0;
}

boolean
  isMutable int qualifiers
{
  = qualifiers & 1 << (MUTABLE - EXTERN) != 0;
}

boolean
  isPrivate int qualifiers
{
  = qualifiers & 1 << (PRIVATE - EXTERN) != 0;
}

boolean
  isProtected int qualifiers
{
  = qualifiers & 1 << (PROTECTED - EXTERN) != 0;
}

boolean
  isPublic int qualifiers
{
  = qualifiers & 1 << (PUBLIC - EXTERN) != 0;
}

boolean
  isStatic int qualifiers
{
  = qualifiers & 1 << (STATIC - EXTERN) != 0;
}

boolean
  isRedeclare int qualifiers
{
  = qualifiers & 1 << (REDECLARE - EXTERN) != 0;
}

boolean
  isRedefine int qualifiers
{
  = qualifiers & 1 << (REDEFINE - EXTERN) != 0;
}

boolean
  isDeferred int qualifiers
{
  = qualifiers & 1 << (DEFERRED - EXTERN) != 0;
}

<doc> Return the {qualifiers} with the redeclare bit set or reset,
    according to {yes_p}.  </doc>
int
	  set int qualifiers
  isRedeclare boolean yes_p
{
  if (yes_p)
    = qualifiers | (1 << (REDECLARE - EXTERN));
  else
    = qualifiers & ~(1 << (REDECLARE - EXTERN));
}

end;

implementation instance
Qualified
{
  <doc> The qualifiers on this object.  </doc>
  public int qualifiers;
}

<doc> Designated initializer.  </doc>
id (self)
  initWithQualifiers int q
{
  qualifiers = q;
}

/******************** Qualifier testing ********************/
<doc> <h4>Qualifier testing</h4> </doc>

boolean
  isExtern
{
  = qualifiers & 1 << (EXTERN - EXTERN) != 0;
}

boolean
  isLocal
{
  = qualifiers & 1 << (LOCAL - EXTERN) != 0;
}

boolean
  isMutable
{
  = qualifiers & 1 << (MUTABLE - EXTERN) != 0;
}

boolean
  isPrivate
{
  = qualifiers & 1 << (PRIVATE - EXTERN) != 0;
}

boolean
  isProtected
{
  = qualifiers & 1 << (PROTECTED - EXTERN) != 0;
}

boolean
  isPublic
{
  = qualifiers & 1 << (PUBLIC - EXTERN) != 0;
}

boolean
  isStatic
{
  = qualifiers & 1 << (STATIC - EXTERN) != 0;
}

boolean
  isRedeclare
{
  = qualifiers & 1 << (REDECLARE - EXTERN) != 0;
}

boolean
  isRedefine
{
  = qualifiers & 1 << (REDEFINE - EXTERN) != 0;
}

boolean
  isDeferred
{
  = qualifiers & 1 << (DEFERRED - EXTERN) != 0;
}

end;

/******************** Reference ********************/

implementation class
Reference: Expression

end;

implementation instance
Reference
{
  <doc> The variable being referenced.  </doc>
  public Variable referenced;
}

<doc> Designated initializer.  </doc>
id
  initWithReferenced Variable v
	fromInstance Instance m
{
  referenced = v;

  Type t = [v type];
  if (m != nil && t != nil)
    t = [t asSeenFromInstance m];

  = [super initWithType t];
}

<doc> Emit the {referenced}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" ref=", referenced)];
}

end;

/******************** StateAttribute ********************/

implementation class
StateAttribute: Attribute, VariableSingle

end;

implementation instance
StateAttribute

<doc> Designated initializer.  </doc>
id
  initWithName String n
	  type Type t
     extension mu.Extension e
{
  [super initWithExtension e];
  = [super initWithName n type t];
}

end;

/******************** Tuple ********************/

implementation class
Tuple: State, MutableObjectArray
end;

implementation instance
Tuple

<doc> Return {YES}.  </doc>
redefine boolean
  isTuple
{
  = YES;
}

<doc> Return the number of non-tuple elements in this tuple.  </doc>
int (n)
  numFlatElements
{
  int i;
  for (i = 0; i < length; i++)
    n += [self[i] numFlatElements];
}

OutputStream
  writeFields OutputStream s
{
  int i;

  [[super (MutableObjectArray) writeFields s] print (" elts=(")];
  while (i < length)
    {
      if (i > 0)
	[s print ", "];
      All elt = self[i++];
      [s print elt];
    }
  = [s print ')'];
}

<doc><h4>Garbage collection</h4></doc>

// Maybe obect arrays should be nice and also invoke super's (State that
// is) implementation of the gc_mark_elements?  Then all subclasses of
// [Mutable]ObjectArray would not need their own...
<doc> Also let the {State} in us mark the other instance variables.
    </doc>
void
  gc_mark_elements
{
  [super (MutableObjectArray) gc_mark_elements];
  [super (State) gc_mark_elements];
}

end;

/******************** Type ********************/

implementation class
Type: Common
end;

implementation instance
Type

<doc> Return the type as seen from the instance {i}.  Most important task
    of this method is to return {id} if the class specified {instance
    (id)}.  The default implementation simply returns {self}.  </doc>
id
  asSeenFromInstance Instance i
{
  = self;
}

<doc> Return {TRUE} if the receiving type is equal to the {other} tuple
    type.  The default implementation returns {FALSE}.  </doc>
boolean
  equalTypeTuple TypeTuple other
{
}

<doc> Return the instance of the receiving object type.  The default
    implementation issues an error and returns {self}.  </doc>
Type
  instance
{
  [[[Parser error] print "request instance of improper type"] nl];
  = self;
}

<doc> Return {TRUE} if this is a basic type.  The default implementation
    returns {FALSE}.  </doc>
boolean
  isBasicType
{
  = NO;
}

<doc> Return {TRUE} if this is an object type.  The default implementation
    returns {FALSE}.  </doc>
boolean
  isMeta
{
  = NO;
}

<doc> Return the meta (i.e. class) of the receiving object type.  The
    default implementation issues an error and returns {self}.  </doc>
Type
  meta
{
  [[[Parser error] print "request class of improper type"] nl];
  = self;
}

<doc> Return the name of the type.  </doc>
deferred String
  name;

end;

/******************** Typed ********************/

<doc> An {Typed} is something with a type.  </doc>
implementation class
Typed: Common
end;

implementation instance
Typed
{
  <doc> The type of this argument.  </doc>
  public mutable Type type;
}

<doc> Designated initializer.  </doc>
id (self)
  initWithType Type t
{
  type = t;
}

<doc> Return the {type}, forcing it to {error_type} if it is {nil}.  </doc>
Type
  force_type
{
  if (!type)
    [[[Parser warning self] print "returning the error-type"] nl];
  = !type ? error_type : type;
}

<doc> Emit the {type}, if it is not {nil}.  </doc>
OutputStream
  writeFields OutputStream s
{
  s = [super writeFields s];
  = !type ? s : [s print (" type=", type)];
}

end;

/******************** TypeSingle ********************/

implementation class
TypeSingle: Named, Type
end;

implementation instance
TypeSingle

end;

/******************** TypeTuple ********************/

implementation class
TypeTuple: Type, Tuple
end;

implementation instance
TypeTuple

<doc> Tell the {other} to {equalTypeTuple}.  </doc>
boolean
  equal Type other
{
  = [other equalTypeTuple self];
}

<doc> Pass to the {equal} of {super}.  </doc>
boolean
  equalTypeTuple TypeTuple other
{
  = [super equal other];
}

// Pity we can't direct this straight to the Tuple...
// Fri Apr 24 14:43:42 1998, tiggr@gerbil.org
<doc> Return {YES}.  </doc>
boolean
  isTuple
{
  = YES;
}

// Needed?
// Mon Apr 27 11:58:36 1998, tiggr@gerbil.org
<doc> A tuple does not have a name.  </doc>
String
  name
{
  [self shouldNotImplement cmd];
}

end;

/******************** Variable ********************/

implementation class
Variable: Typed
end;

implementation instance
Variable

end;

/******************** VariableSingle ********************/

implementation class
VariableSingle: Variable, Named

end;

implementation instance
VariableSingle

<doc> Designated initializer.  </doc>
id
  initWithName String n
	  type Type t
{
  [super initWithType t];
  = [super initWithName n];
}

<doc> If an argument to the method {m} with the same name as the receiving
    {Variable} exists, return that argument, otherwise return {self}.
    </doc>
Named (retval)
  resolveReturnNamesInMethod CustomMethod m
{
  retval = [m argumentNamed name];
  if (!retval)
    retval = self;
}

<doc> Invoke the {writeFields} of {Typed}, but do the name printing self,
    to avoid {Common} doing its thing twice.  </doc>
<doc> Emit the {name}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super (Variable) writeFields s] print (" name=", name)];
}

end;

/******************** VariableTuple ********************/

implementation class
VariableTuple: Variable, Tuple
end;

implementation instance
VariableTuple

// Pity we can't direct this straight to the Tuple...
// Fri Apr 24 14:43:42 1998, tiggr@gerbil.org
<doc> Return {YES}.  </doc>
boolean
  isTuple
{
  = YES;
}

<doc> Direct at the {Tuple}.  </doc>
int
  numFlatElements
{
  = [super (Tuple) numFlatElements];
}


<doc> Direct at the {Tuple}.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [super (Tuple) writeFields s];
}

end;
